from pyqrllib.pyqrllib import bin2hstr

from qrl.core.State import State
from qrl.core.StateContainer import StateContainer
from qrl.core.VoteStats import VoteStats
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.MultiSigAddressState import MultiSigAddressState
from qrl.core.misc import logger
from qrl.core.txs.Transaction import Transaction
from qrl.crypto.misc import sha256


class MultiSigSpend(Transaction):
    """
    MultiSigSpend for the transaction of QRL from a multi sig wallet to another wallet.
    """

    def __init__(self, protobuf_transaction=None):
        super(MultiSigSpend, self).__init__(protobuf_transaction)

    @property
    def multi_sig_address(self):
        return self._data.multi_sig_spend.multi_sig_address

    @property
    def addrs_to(self):
        return self._data.multi_sig_spend.addrs_to

    @property
    def total_amount(self):
        total_amount = 0
        for amount in self.amounts:
            total_amount += amount
        return total_amount

    @property
    def amounts(self):
        return self._data.multi_sig_spend.amounts

    @property
    def expiry_block_number(self):
        return self._data.multi_sig_spend.expiry_block_number

    def get_data_hash(self):
        tmp_tx_hash = (self.master_addr +
                       self.fee.to_bytes(8, byteorder='big', signed=False) +
                       self.multi_sig_address +
                       self.expiry_block_number.to_bytes(8, byteorder='big', signed=False))

        for index in range(0, len(self.addrs_to)):
            tmp_tx_hash = (tmp_tx_hash +
                           self.addrs_to[index] +
                           self.amounts[index].to_bytes(8, byteorder='big', signed=False))

        return sha256(tmp_tx_hash)

    @staticmethod
    def create(multi_sig_address: bytes,
               addrs_to: list,
               amounts: list,
               expiry_block_number: int,
               fee: int,
               xmss_pk,
               master_addr: bytes = None):
        multi_sig_spend = MultiSigSpend()

        if master_addr:
            multi_sig_spend._data.master_addr = master_addr

        multi_sig_spend._data.public_key = bytes(xmss_pk)

        multi_sig_spend._data.multi_sig_spend.multi_sig_address = multi_sig_address

        for addr_to in addrs_to:
            multi_sig_spend._data.multi_sig_spend.addrs_to.append(addr_to)

        for amount in amounts:
            multi_sig_spend._data.multi_sig_spend.amounts.append(amount)

        multi_sig_spend._data.multi_sig_spend.expiry_block_number = expiry_block_number

        multi_sig_spend._data.fee = int(fee)

        multi_sig_spend.validate_or_raise(verify_signature=False)

        return multi_sig_spend

    def _validate_custom(self):
        for amount in self.amounts:
            if amount == 0:
                logger.warning('Amount cannot be 0 - %s', self.amounts)
                logger.warning('Invalid TransferTransaction')
                return False

        if self.fee < 0:
            logger.warning('MultiSigSpend [%s] Invalid Fee = %d', bin2hstr(self.txhash), self.fee)
            return False

        if len(self.addrs_to) == 0:
            logger.warning('[MultiSigSpend] No addrs_to found')
            return False

        if len(self.addrs_to) != len(self.amounts):
            logger.warning('[MultiSigSpend] Mismatch number of addresses to & amounts')
            logger.warning('>> Length of addrs_to %s', len(self.addrs_to))
            logger.warning('>> Length of amounts %s', len(self.amounts))
            return False

        if not MultiSigAddressState.address_is_valid(self.multi_sig_address):
            logger.warning('[MultiSigSpend] Invalid MultiSig Address')
            logger.warning('Multi Sig Address %s', self.multi_sig_address)
            return False

        if not OptimizedAddressState.address_is_valid(self.addr_from):
            logger.warning('[MultiSigSpend] Invalid address addr_from: %s', bin2hstr(self.addr_from))
            return False

        for addr_to in self.addrs_to:
            if not (OptimizedAddressState.address_is_valid(addr_to) or MultiSigAddressState.address_is_valid(addr_to)):
                logger.warning('[MultiSigSpend] Invalid address addr_to: %s', bin2hstr(addr_to))
                return False

        return True

    def _validate_extended(self, state_container: StateContainer):
        if state_container.block_number < state_container.current_dev_config.hard_fork_heights[0]:
            logger.warning("[MultiSigSpend] Hard Fork Feature not yet activated")
            return False

        if len(self.addrs_to) > state_container.current_dev_config.transaction_multi_output_limit:
            logger.warning('[MultiSigSpend] Number of addresses exceeds max limit')
            logger.warning('Number of addresses %s', len(self.addrs_to))
            logger.warning('Number of amounts %s', len(self.amounts))
            return False

        addr_from_state = state_container.addresses_state[self.addr_from]
        if self.multi_sig_address not in state_container.addresses_state:
            logger.error("[MultiSigSpend] Multi Sig address state not found in state_container %s",
                         self.multi_sig_address)
            return False

        multi_sig_address_state = state_container.addresses_state[self.multi_sig_address]
        block_number = state_container.block_number

        if addr_from_state.address != self.addr_from:
            logger.error("[MultiSigSpend] Unexpected addr_from_state")
            logger.error("Expecting State for address %s, but got state for address %s",
                         bin2hstr(self.addr_from),
                         bin2hstr(addr_from_state.address))
            return False

        if multi_sig_address_state.address != self.multi_sig_address:
            logger.error("[MultiSigSpend] Unexpected multi sig address state")
            logger.error("Expecting State for address %s, but got state for address %s",
                         bin2hstr(self.multi_sig_address),
                         bin2hstr(multi_sig_address_state.address))
            return False

        tx_balance = addr_from_state.balance
        total_amount = self.total_amount

        if tx_balance < self.fee:
            logger.info('[MultiSigSpend] State validation failed for %s because: Insufficient funds',
                        bin2hstr(self.txhash))
            logger.info('address: %s, balance: %s, fee: %s', bin2hstr(self.addr_from), tx_balance, self.fee)
            return False

        if multi_sig_address_state.balance < total_amount:
            logger.info('[MultiSigSpend] State validation failed for %s because: Insufficient funds',
                        bin2hstr(self.txhash))
            logger.info('address: %s, balance: %s, fee: %s', bin2hstr(self.multi_sig_address), tx_balance, self.fee)
            return False

        # Multi Sig Spend considered to be expired after block having block number equals to
        # self.expiry_block_number gets added into the main chain
        if self.expiry_block_number <= block_number:
            logger.info('[MultiSigSpend] State validation failed for %s due to invalid expiry_block_number',
                        bin2hstr(self.txhash))
            logger.info('Chain Height: %s, Expiry Block Number: %s',
                        block_number,
                        self.expiry_block_number)
            return False

        if self.addr_from not in multi_sig_address_state.signatories:
            logger.info('[MultiSigSpend] Address is not in the signatories list: %s',
                        bin2hstr(self.addr_from))
            return False

        return True

    def set_affected_address(self, addresses_set: set):
        super().set_affected_address(addresses_set)
        addresses_set.add(self.multi_sig_address)
        for addrs_to in self.addrs_to:
            addresses_set.add(addrs_to)

    def apply(self,
              state: State,
              state_container: StateContainer) -> bool:
        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container, self.fee, subtract=True)
        state_container.paginated_tx_hash.insert(address_state, self.txhash)

        # TODO: Following line might not be needed
        state_container.multi_sig_spend_txs[self.txhash] = self

        multi_sig_address_state = state_container.addresses_state[self.multi_sig_address]
        for signatory_address in multi_sig_address_state.signatories:
            signatory_address_state = state_container.addresses_state[signatory_address]
            state_container.paginated_multi_sig_spend.insert(signatory_address_state, self.txhash)

        state_container.paginated_tx_hash.insert(multi_sig_address_state, self.txhash)
        state_container.paginated_multi_sig_spend.insert(multi_sig_address_state, self.txhash)

        vote_stats = VoteStats.create(self.multi_sig_address,
                                      self.txhash,
                                      state_container.addresses_state[self.multi_sig_address].signatories,
                                      self.expiry_block_number)
        state_container.votes_stats[self.txhash] = vote_stats

        return self._apply_state_changes_for_PK(state_container)

    def revert(self,
               state: State,
               state_container: StateContainer) -> bool:
        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container, self.fee)
        state_container.paginated_tx_hash.remove(address_state, self.txhash)

        multi_sig_address_state = state_container.addresses_state[self.multi_sig_address]
        for signatory_address in multi_sig_address_state.signatories:
            signatory_address_state = state_container.addresses_state[signatory_address]
            state_container.paginated_multi_sig_spend.remove(signatory_address_state, self.txhash)

        state_container.paginated_tx_hash.remove(multi_sig_address_state, self.txhash)
        state_container.paginated_multi_sig_spend.remove(multi_sig_address_state, self.txhash)

        VoteStats.delete_state(state, self.txhash, state_container.batch)

        if self.txhash in state_container.votes_stats:
            del state_container.votes_stats[self.txhash]

        return self._revert_state_changes_for_PK(state_container)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase
from mock import MagicMock

from qrl.core.LastTransactions import LastTransactions
from qrl.core.misc import logger, db
from qrl.core.State import State
from qrl.core.txs.TransferTransaction import TransferTransaction
from qrl.core.Block import Block

from tests.misc.helper import set_qrl_dir, get_alice_xmss, get_some_address

logger.initialize_default()


class TestLastTransactions(TestCase):
    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()
        self.m_db = MagicMock(name='mock DB', autospec=db.DB)

    def test_update_last_tx(self):
        alice_xmss = get_alice_xmss()
        # Test Case: When there is no last txns
        self.assertEqual(LastTransactions.get_last_txs(self.state), [])

        block = Block()
        tx1 = TransferTransaction.create(addrs_to=[get_some_address(1), get_some_address(2)],
                                         amounts=[1, 2],
                                         message_data=None,
                                         fee=0,
                                         xmss_pk=alice_xmss.pk)
        block._data.transactions.extend([tx1.pbdata])
        LastTransactions._update_last_tx(self.state, block, None)
        last_txns = LastTransactions.get_last_txs(self.state)

        # Test Case: When there is only 1 last txns
        self.assertEqual(len(last_txns), 1)
        self.assertEqual(last_txns[0].to_json(), tx1.to_json())

        block = Block()
        tx2 = TransferTransaction.create(addrs_to=[get_some_address(2), get_some_address(3)],
                                         amounts=[1, 2],
                                         message_data=None,
                                         fee=0,
                                         xmss_pk=alice_xmss.pk)

        tx3 = TransferTransaction.create(addrs_to=[get_some_address(4), get_some_address(5)],
                                         amounts=[1, 2],
                                         message_data=None,
                                         fee=0,
                                         xmss_pk=alice_xmss.pk)
        block._data.transactions.extend([tx2.pbdata, tx3.pbdata])
        LastTransactions._update_last_tx(self.state, block, None)
        last_txns = LastTransactions.get_last_txs(self.state)

        # Test Case: When there are 3 last txns
        self.assertEqual(len(last_txns), 3)
        self.assertEqual(last_txns[0].to_json(),
                         tx3.to_json())
        self.assertEqual(last_txns[1].to_json(),
                         tx2.to_json())
        self.assertEqual(last_txns[2].to_json(),
                         tx1.to_json())

    def test_get_last_txs(self):
        self.assertEqual(LastTransactions.get_last_txs(self.state), [])

        alice_xmss = get_alice_xmss()
        block = Block()
        tx1 = TransferTransaction.create(addrs_to=[get_some_address(0), get_some_address(1)],
                                         amounts=[1, 2],
                                         message_data=None,
                                         fee=0,
                                         xmss_pk=alice_xmss.pk)
        block._data.transactions.extend([tx1.pbdata])
        LastTransactions._update_last_tx(self.state, block, None)
        last_txns = LastTransactions.get_last_txs(self.state)

        # Test Case: When there is only 1 last txns
        self.assertEqual(len(last_txns), 1)
        self.assertEqual(last_txns[0].to_json(), tx1.to_json())

    def test_remove_last_tx(self):
        # Test Case: When there is no last txns
        self.assertEqual(LastTransactions.get_last_txs(self.state), [])

        alice_xmss = get_alice_xmss()

        block = Block()
        tx1 = TransferTransaction.create(addrs_to=[get_some_address(1), get_some_address(2)],
                                         amounts=[1, 2],
                                         message_data=None,
                                         fee=0,
                                         xmss_pk=alice_xmss.pk)
        block._data.transactions.extend([tx1.pbdata])
        LastTransactions._update_last_tx(self.state, block, None)
        last_txns = LastTransactions.get_last_txs(self.state)

        self.assertEqual(last_txns[0].to_json(), tx1.to_json())

        LastTransactions._remove_last_tx(self.state, block, None)
        last_txns = LastTransactions.get_last_txs(self.state)
        self.assertEqual(last_txns, [])



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase
import mock
from mock import MagicMock, patch

from qrl.core import config
from qrl.core.misc import logger, db
from qrl.core.State import State
from qrl.core.Block import Block
from qrl.generated import qrlstateinfo_pb2

from tests.misc.helper import set_qrl_dir, get_alice_xmss, replacement_getTime, gen_blocks

logger.initialize_default()


@mock.patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestState(TestCase):
    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()
        self.m_db = MagicMock(name='mock DB', autospec=db.DB)

    def test_create_state(self):
        self.assertIsNotNone(self.state)  # to avoid warning (unused variable)

    def test_release_state(self):
        self.assertIsNotNone(self.state)  # to avoid warning (unused variable)

    def test_basic_state_funcs(self):
        batch = self.state.batch
        self.assertIsNotNone(batch)
        self.state.write_batch(batch)
        self.assertEqual(self.state.total_coin_supply, 0)

    def test_address_used(self):
        alice_xmss = get_alice_xmss()
        self.assertFalse(self.state.get_address_is_used(alice_xmss.address))

    def test_get_batch(self):
        self.assertIsNotNone(self.state.batch)

    def test_write_batch(self):
        batch = self.state.batch
        block = Block.create(dev_config=config.dev,
                             block_number=10,
                             prev_headerhash=b'aa',
                             prev_timestamp=10,
                             transactions=[],
                             miner_address=b'aa',
                             seed_height=0,
                             seed_hash=None)
        Block.put_block(self.state, block, batch)
        self.assertIsNone(Block.get_block(self.state, block.headerhash))
        self.state.write_batch(batch)
        block2 = Block.get_block(self.state, block.headerhash)
        self.assertEqual(block.headerhash, block2.headerhash)

    def test_update_total_coin_supply(self):
        self.assertEqual(self.state.total_coin_supply, 0)
        self.state._update_total_coin_supply(100, None)
        self.assertEqual(self.state.total_coin_supply, 100)

    def test_total_coin_supply(self):
        self.assertEqual(self.state.total_coin_supply, 0)

    def test_delete(self):
        block = Block()
        Block.put_block(self.state, block, None)
        block1 = Block.get_block(self.state, block.headerhash)
        self.assertEqual(block.serialize(), block1.serialize())
        self.state._delete(block.headerhash, None)
        self.assertIsNone(Block.get_block(self.state, block.headerhash))

    def test_get_block_size_limit(self):
        alice_xmss = get_alice_xmss()
        blocks = gen_blocks(20, self.state, alice_xmss.address)
        self.assertEqual(Block.get_block_size_limit(self.state, blocks[-1], config.dev), 1048576)

        # get_block_size_limit() should return None if it couldn't get any blocks from db
        with patch('qrl.core.Block.Block.get_block', return_value=None):
            self.assertIsNone(Block.get_block_size_limit(self.state, blocks[-1], config.dev))

    def test_update_mainchain_height(self):
        self.state.update_mainchain_height(5, None)
        self.assertEqual(self.state.get_mainchain_height(), 5)

    def test_get_mainchain_height(self):
        # Test Case: Check default value
        self.assertEqual(self.state.get_mainchain_height(), -1)

        self.state.update_mainchain_height(15, None)
        self.assertEqual(self.state.get_mainchain_height(), 15)

        self.state.update_mainchain_height(5, None)
        self.assertEqual(self.state.get_mainchain_height(), 5)

    def test_fork_state(self):
        fork_state = qrlstateinfo_pb2.ForkState(
            initiator_headerhash=b'block2_right',
            fork_point_headerhash=b'block0_base_of_fork',
            old_mainchain_hash_path=[b'block1_right', b'block2_right'],
            new_mainchain_hash_path=[b'block1_left', b'block2_left']
        )
        self.assertIsNone(self.state.get_fork_state())

        self.state.put_fork_state(fork_state)
        self.assertEqual(fork_state, self.state.get_fork_state())

        self.state.delete_fork_state()
        self.assertIsNone(self.state.get_fork_state())



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from math import ceil

from qrl.core import config
from qrl.core.PaginatedData import PaginatedData
from qrl.core.misc import logger
from qrl.generated import qrl_pb2


class PaginatedBitfield(PaginatedData):
    def __init__(self, write_access: bool, db):
        super(PaginatedBitfield, self).__init__(b'bitfield', write_access, db)

    def generate_bitfield_key(self, address, page):
        return self.name + b'_' + address + b'_' + page.to_bytes(8, byteorder='big', signed=False)

    def load_bitfield_and_ots_key_reuse(self, address, ots_key_index) -> bool:
        page = (ots_key_index // config.dev.ots_tracking_per_page) + 1
        key = self.generate_bitfield_key(address, page)
        self.load_bitfield(address, ots_key_index)
        ots_bitfield = self.key_value[key]

        return self.ots_key_reuse(ots_bitfield, ots_key_index)

    @staticmethod
    def ots_key_reuse(ots_bitfield, ots_key_index) -> bool:
        offset = ots_key_index >> 3
        relative = ots_key_index % 8
        bitfield = bytearray(ots_bitfield[offset])
        bit_value = (bitfield[0] >> relative) & 1

        if bit_value:
            return True

        return False

    def set_ots_key(self, addresses_state: dict, address, ots_key_index):
        page = ots_key_index // config.dev.ots_tracking_per_page + 1
        key = self.generate_bitfield_key(address, page)
        self.load_bitfield(address, ots_key_index)
        ots_bitfield = self.key_value[key]

        ots_key_index = ots_key_index % config.dev.ots_tracking_per_page
        offset = ots_key_index >> 3
        relative = ots_key_index % 8
        bitfield = bytearray(ots_bitfield[offset])
        ots_bitfield[offset] = bytes([bitfield[0] | (1 << relative)])
        address_state = addresses_state[address]
        address_state.used_ots_key_count += 1
        self.update_used_page_in_address_state(address, addresses_state, page)

    def update_used_page_in_address_state(self, address, addresses_state: dict, page: int):
        # TODO: Write Unit Test
        address_state = addresses_state[address]
        if address_state.ots_bitfield_used_page == page - 1:
            remaining_ots = 2 ** address_state.height - (page - 1) * config.dev.ots_tracking_per_page
            while remaining_ots > 0:
                key = self.generate_bitfield_key(address, page)
                if key not in self.key_value:
                    self.key_value[key] = self.get_paginated_data(address, page)
                ots_bitfield = self.key_value[key]

                for i in range(min(config.dev.ots_bitfield_size, ceil(remaining_ots / 8))):
                    if ots_bitfield[i] != b'\xff':
                        return
                    if remaining_ots >= 8:
                        remaining_ots -= 8
                address_state.ots_bitfield_used_page = page  # TODO: Replace by setter function
                page += 1
                if page * config.dev.ots_tracking_per_page > 2 ** address_state.height:
                    return

    def unset_ots_key(self, addresses_state: dict, address, ots_key_index):
        page = ots_key_index // config.dev.ots_tracking_per_page + 1
        key = self.generate_bitfield_key(address, page)
        self.load_bitfield(address, ots_key_index)
        ots_bitfield = self.key_value[key]

        ots_key_index = ots_key_index % config.dev.ots_tracking_per_page
        offset = ots_key_index >> 3
        relative = ots_key_index % 8
        bitfield = bytearray(ots_bitfield[offset])
        ots_bitfield[offset] = bytes([bitfield[0] & ~(1 << relative)])
        address_state = addresses_state[address]
        address_state.used_ots_key_count -= 1
        if address_state.ots_bitfield_used_page >= page:
            address_state.ots_bitfield_used_page = page - 1  # TODO: Replace by setter function

    def load_bitfield(self, address, ots_key_index):
        page = (ots_key_index // config.dev.ots_tracking_per_page) + 1
        key = self.generate_bitfield_key(address, page)
        if key not in self.key_value:
            self.key_value[key] = self.get_paginated_data(address, page)

    def get_paginated_data(self, key, page):
        try:
            pbData = self.db.get_raw(self.name + b'_' + key + b'_' + page.to_bytes(8, byteorder='big', signed=False))
            data_list = qrl_pb2.DataList()
            data_list.ParseFromString(bytes(pbData))
            return list(data_list.values)
        except KeyError:
            return [b'\x00'] * config.dev.ots_bitfield_size
        except Exception as e:
            logger.error('[get_paginated_data] Exception for %s', self.name)
            logger.exception(e)
            raise

    def put_addresses_bitfield(self, batch):
        if not self.write_access:
            return
        for key in self.key_value:
            data_list = qrl_pb2.DataList(values=self.key_value[key])
            self.db.put_raw(key,
                            data_list.SerializeToString(),
                            batch)



from pyqrllib.pyqrllib import bin2hstr, QRLHelper
from typing import Union

from qrl.core.State import State
from qrl.core.StateContainer import StateContainer
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.misc import logger
from qrl.core.txs.Transaction import Transaction


class MessageTransaction(Transaction):

    def __init__(self, protobuf_transaction=None):
        super(MessageTransaction, self).__init__(protobuf_transaction)

    @property
    def message_hash(self):
        return self._data.message.message_hash

    @property
    def addr_to(self):
        return self._data.message.addr_to

    def get_data_bytes(self):
        return self.master_addr + \
               self.fee.to_bytes(8, byteorder='big', signed=False) + \
               self.message_hash + \
               self.addr_to

    @staticmethod
    def create(message_hash: bytes, addr_to: Union[bytes, None], fee: int, xmss_pk: bytes, master_addr: bytes = None):
        transaction = MessageTransaction()

        if master_addr:
            transaction._data.master_addr = master_addr

        transaction._data.message.message_hash = message_hash
        if addr_to:
            transaction._data.message.addr_to = addr_to

        transaction._data.fee = fee

        transaction._data.public_key = xmss_pk

        transaction.validate_or_raise(verify_signature=False)

        return transaction

    def _validate_custom(self) -> bool:
        if len(self.message_hash) == 0:
            logger.warning('Message cannot be empty')
            return False

        if len(self.addr_to) > 0 and not (OptimizedAddressState.address_is_valid(self.addr_to)):
            logger.warning('[MessageTransaction] Invalid address addr_to: %s', bin2hstr(self.addr_to))
            return False

        if self.fee < 0:
            logger.info('State validation failed for %s because: Negative send', bin2hstr(self.txhash))
            return False

        return True

    def _validate_extended(self, state_container: StateContainer) -> bool:
        if len(self.addr_to) != 0:
            if state_container.block_number < state_container.current_dev_config.hard_fork_heights[0]:
                logger.warning("[MessageTransaction] Hard Fork Feature not yet activated")
                return False

        if len(self.message_hash) > state_container.current_dev_config.message_max_length:  # TODO: Move to dev config
            logger.warning('Message length cannot be more than %s', state_container.current_dev_config.message_max_length)
            logger.warning('Found message length %s', len(self.message_hash))
            return False

        tx_balance = state_container.addresses_state[self.addr_from].balance

        if tx_balance < self.fee:
            logger.info('State validation failed for %s because: Insufficient funds', bin2hstr(self.txhash))
            logger.info('balance: %s, amount: %s', tx_balance, self.fee)
            return False

        return True

    def set_affected_address(self, addresses_set: set):
        super().set_affected_address(addresses_set)
        if self.addr_to:
            addresses_set.add(self.addr_to)

    def apply(self,
              state: State,
              state_container: StateContainer) -> bool:
        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container, self.fee, subtract=True)
        state_container.paginated_tx_hash.insert(address_state, self.txhash)

        if self.addr_to:
            addr_from_pk = bytes(QRLHelper.getAddress(self.PK))
            address_state = state_container.addresses_state[self.addr_to]
            if self.addr_to not in (self.addr_from, addr_from_pk):
                state_container.paginated_tx_hash.insert(address_state, self.txhash)
                state_container.paginated_inbox_message.insert(address_state, self.txhash)

        return self._apply_state_changes_for_PK(state_container)

    def revert(self,
               state: State,
               state_container: StateContainer) -> bool:
        address_state = state_container.addresses_state[self.addr_from]
        address_state.update_balance(state_container, self.fee)
        state_container.paginated_tx_hash.remove(address_state, self.txhash)

        if self.addr_to:
            addr_from_pk = bytes(QRLHelper.getAddress(self.PK))
            address_state = state_container.addresses_state[self.addr_to]
            if self.addr_to not in (self.addr_from, addr_from_pk):
                state_container.paginated_tx_hash.remove(address_state, self.txhash)
                state_container.paginated_inbox_message.remove(address_state, self.txhash)

        return self._revert_state_changes_for_PK(state_container)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from abc import ABCMeta, abstractmethod
from math import log, floor

from google.protobuf.json_format import MessageToJson, Parse
from pyqrllib.pyqrllib import bin2hstr, QRLHelper, XmssFast

from qrl.core import config
from qrl.core.State import State
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.StateContainer import StateContainer
from qrl.core.misc import logger
from qrl.core.txs import build_tx as main_build_tx
from qrl.crypto.misc import sha256
from qrl.generated import qrl_pb2

CODEMAP = {
    'transfer': 1,
    'coinbase': 2,
    'latticePK': 3,
    'message': 4,
    'token': 5,
    'transfer_token': 6,
    'slave': 7,
    'multi_sig_create': 8,
    'multi_sig_spend': 9,
    'multi_sig_vote': 10,
}


class Transaction(object, metaclass=ABCMeta):
    """
    Abstract Base class to be derived by all other transactions
    """

    def __init__(self, protobuf_transaction=None):
        self._data = protobuf_transaction  # This object cointains persistable data
        if protobuf_transaction is None:
            self._data = qrl_pb2.Transaction()

    def __lt__(self, tx):
        if self.fee < tx.fee:
            return True
        return False

    def __gt__(self, tx):
        if self.fee > tx.fee:
            return True
        return False

    @property
    def size(self):
        return self._data.ByteSize()

    @property
    def pbdata(self):
        """
        Returns a protobuf object that contains persistable data representing this object
        :return: A protobuf Transaction object
        :rtype: qrl_pb2.Transaction
        """
        return self._data

    @property
    def type(self):
        return self._data.WhichOneof('transactionType')

    @property
    def fee(self):
        return self._data.fee

    @property
    def nonce(self):
        return self._data.nonce

    @property
    def master_addr(self):
        return self._data.master_addr

    @property
    def addr_from(self):
        if self.master_addr:
            return self.master_addr

        return bytes(QRLHelper.getAddress(self.PK))

    @property
    def ots_key(self):
        return self.get_ots_from_signature(self.signature)

    @staticmethod
    def get_ots_from_signature(signature):
        try:
            return int(bin2hstr(signature)[0:8], 16)
        except ValueError:
            raise ValueError('OTS Key Index: First 4 bytes of signature are invalid')

    @staticmethod
    def calc_allowed_decimals(value):
        if not isinstance(value, int):
            raise ValueError('value should be of integer type')
        if value == 0:
            return 19

        # floor value could be negative, so return 0 when the floor value is negative
        return max(floor(19 - log(value, 10)), 0)

    @property
    def PK(self):
        return self._data.public_key

    @property
    def signature(self):
        return self._data.signature

    @staticmethod
    def from_pbdata(pbdata: qrl_pb2.Transaction):
        return main_build_tx(pbdata.WhichOneof('transactionType'), pbdata)

    @staticmethod
    def from_json(json_data):
        pbdata = qrl_pb2.Transaction()
        Parse(json_data, pbdata)
        return Transaction.from_pbdata(pbdata)

    @staticmethod
    def get_slave(tx):
        addr_from_pk = bytes(QRLHelper.getAddress(tx.PK))
        if addr_from_pk != tx.addr_from:
            return addr_from_pk
        return None

    @property
    def txhash(self) -> bytes:
        return self._data.transaction_hash

    def update_txhash(self):
        self._data.transaction_hash = self.generate_txhash()

    def generate_txhash(self):
        return sha256(
            self.get_data_hash() +
            self.signature +
            self.PK
        )

    def get_data_bytes(self) -> bytes:
        """
        This method returns the essential bytes that represent the transaction and will be later signed
        :return:
        """
        raise NotImplementedError

    def get_data_hash(self) -> bytes:
        """
        This method returns the hashes of the transaction data.
        :return:
        """
        return sha256(self.get_data_bytes())

    def sign(self, object_with_sign_method):
        self._data.signature = object_with_sign_method.sign(self.get_data_hash())
        self.update_txhash()

    def set_affected_address(self, addresses_set: set):
        addresses_set.add(self.addr_from)
        addresses_set.add(bytes(QRLHelper.getAddress(self.PK)))

    @abstractmethod
    def _validate_custom(self) -> bool:
        """
        This is an extension point for derived classes validation
        If derived classes need additional field validation they should override this member
        """
        raise NotImplementedError

    @abstractmethod
    def _validate_extended(self, state_container: StateContainer) -> bool:
        raise NotImplementedError

    @abstractmethod
    def apply(self,
              state: State,
              state_container: StateContainer) -> bool:
        raise NotImplementedError

    @abstractmethod
    def revert(self,
               state: State,
               state_container: StateContainer) -> bool:
        raise NotImplementedError

    def validate_transaction_pool(self, transaction_pool):
        for tx_set in transaction_pool:
            txn = tx_set[1].transaction
            if txn.txhash == self.txhash:
                continue

            if self.PK != txn.PK:
                continue

            if txn.ots_key == self.ots_key:
                logger.info('State validation failed for %s because: OTS Public key re-use detected',
                            bin2hstr(self.txhash))
                logger.info('Subtype %s', self.type)
                return False

        return True

    def validate(self, verify_signature=True) -> bool:
        """
        This method calls validate_or_raise, logs any failure and returns True or False accordingly
        The main purpose is to avoid exceptions and accommodate legacy code
        :return: True is the transaction is valid
        :rtype: bool
        """
        try:
            self.validate_or_raise(verify_signature)
        except ValueError as e:
            logger.info('[%s] failed validate_tx', bin2hstr(self.txhash))
            logger.warning(str(e))
            return False
        except Exception as e:
            logger.exception(e)
            return False
        return True

    def validate_all(self, state_container: StateContainer, check_nonce=True) -> bool:
        if self.pbdata.WhichOneof('transactionType') == 'coinbase':
            if not self._validate_extended(state_container):
                return False
            return True

        if not self.validate(True):  # It also calls _validate_custom
            return False
        if not self.validate_slave(state_container):
            return False
        if not self._validate_extended(state_container):
            return False

        addr_from_pk = bytes(QRLHelper.getAddress(self.PK))
        addr_from_pk_state = state_container.addresses_state[addr_from_pk]

        expected_nonce = addr_from_pk_state.nonce + 1

        if check_nonce and self.nonce != expected_nonce:
            logger.warning('nonce incorrect, invalid tx')
            logger.warning('subtype: %s', self.type)
            logger.warning('%s actual: %s expected: %s',
                           OptimizedAddressState.bin_to_qaddress(addr_from_pk),
                           self.nonce,
                           expected_nonce)
            return False

        if state_container.paginated_bitfield.load_bitfield_and_ots_key_reuse(addr_from_pk_state.address,
                                                                              self.ots_key):
            logger.warning('pubkey reuse detected: invalid tx %s', bin2hstr(self.txhash))
            logger.warning('subtype: %s', self.type)
            return False

        return True

    # TODO: will need state_container
    def _coinbase_filter(self):
        if config.dev.coinbase_address in [bytes(QRLHelper.getAddress(self.PK)), self.master_addr]:
            raise ValueError('Coinbase Address only allowed to do Coinbase Transaction')

    def _get_allowed_access_types(self):
        return [0]

    def _get_master_address(self):
        return self.addr_from

    # TODO: will need state_container
    def validate_or_raise(self, verify_signature=True) -> bool:
        """
        This method will validate a transaction and raise exception if problems are found
        :return: True if the exception is valid, exceptions otherwise
        :rtype: bool
        """
        if not self._validate_custom():
            raise ValueError("Custom validation failed")

        self._coinbase_filter()

        expected_transaction_hash = self.generate_txhash()
        if verify_signature and self.txhash != expected_transaction_hash:
            logger.warning('Invalid Transaction hash')
            logger.warning('Expected Transaction hash %s', bin2hstr(expected_transaction_hash))
            logger.warning('Found Transaction hash %s', bin2hstr(self.txhash))
            raise ValueError("Invalid Transaction Hash")

        if verify_signature:
            # Temporarily disabled following new added lines.
            # TODO: Review Juan
            # if not XMSS.validate_signature(self.signature, self.PK):
            #     raise ValueError("Invalid xmss signature")

            if not XmssFast.verify(self.get_data_hash(),
                                   self.signature,
                                   self.PK):
                raise ValueError("Invalid xmss signature")

        return True

    def validate_slave(self, state_container: StateContainer) -> bool:
        addr_from_pk = bytes(QRLHelper.getAddress(self.PK))

        master_address = self._get_master_address()
        allowed_access_types = self._get_allowed_access_types()

        if self.master_addr == addr_from_pk:
            logger.warning('Matching master_addr field and address from PK')
            return False

        if addr_from_pk != master_address:
            if (self.addr_from, self.PK) not in state_container.slaves.data:
                logger.warning("Public key and address doesn't match")
                return False

            slave_access_type = state_container.slaves.data[(self.addr_from, self.PK)].access_type
            if slave_access_type not in allowed_access_types:
                logger.warning('Access Type %s', slave_access_type)
                logger.warning('Slave Address doesnt have sufficient permission')
                return False

        return True

    def get_message_hash(self):
        # FIXME: refactor, review that things are not recalculated too often, cache, etc.
        return self.txhash

    def to_json(self):
        # FIXME: Remove once we move completely to protobuf
        return MessageToJson(self._data, sort_keys=True)

    def serialize(self) -> str:
        return self._data.SerializeToString()

    @staticmethod
    def deserialize(data):
        pbdata = qrl_pb2.Transaction()
        pbdata.ParseFromString(bytes(data))
        tx = Transaction(pbdata)
        return tx

    def _apply_state_changes_for_PK(self,
                                    state_container: StateContainer) -> bool:
        addr_from_pk = bytes(QRLHelper.getAddress(self.PK))
        address_state = state_container.addresses_state[addr_from_pk]
        if self.addr_from != addr_from_pk:
            state_container.paginated_tx_hash.insert(address_state, self.txhash)
        address_state.increase_nonce()
        state_container.paginated_bitfield.set_ots_key(state_container.addresses_state, addr_from_pk, self.ots_key)

        return True

    def _revert_state_changes_for_PK(self,
                                     state_container: StateContainer) -> bool:
        addr_from_pk = bytes(QRLHelper.getAddress(self.PK))
        address_state = state_container.addresses_state[addr_from_pk]
        if self.addr_from != addr_from_pk:
            state_container.paginated_tx_hash.remove(address_state, self.txhash)
        address_state.decrease_nonce()
        state_container.paginated_bitfield.unset_ots_key(state_container.addresses_state, addr_from_pk, self.ots_key)

        return True



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import copy
import threading
from typing import Optional

from pyqrllib.pyqrllib import bin2hstr, hstr2bin
from pyqryptonight.pyqryptonight import UInt256ToString

from qrl.core.config import DevConfig
from qrl.core.miners.qryptonight7.CNv1Miner import CNv1Miner
from qrl.core.miners.qrandomx.QRXMiner import QRandomXMiner
from qrl.core.OptimizedAddressState import OptimizedAddressState
from qrl.core.Block import Block
from qrl.core.DifficultyTracker import DifficultyTracker
from qrl.core.TransactionPool import TransactionPool
from qrl.core.misc import logger
from qrl.core.txs.Transaction import Transaction
from qrl.crypto.Qryptonight import Qryptonight


class Miner:
    def __init__(self,
                 chain_manager,
                 pre_block_logic,
                 mining_address: bytes,
                 mining_thread_count):
        self.lock = threading.RLock()

        self.qryptonight_7_miner = CNv1Miner(pre_block_logic,
                                             mining_address,
                                             mining_thread_count,
                                             self.lock)

        self.qrandomx_miner = QRandomXMiner(chain_manager,
                                            pre_block_logic,
                                            mining_address,
                                            mining_thread_count,
                                            self.lock)
        self._qn = Qryptonight()
        self._chain_manager = chain_manager
        self._pre_block_logic = pre_block_logic
        self._mining_block = None
        self._current_difficulty = None
        self._current_target = None
        self._measurement = None

        self._current_miner = self.qrandomx_miner

    def isRunning(self):
        return self._current_miner.isRunning

    def get_miner(self, height, dev_config: DevConfig):
        if height < dev_config.hard_fork_heights[0]:
            self._current_miner = self.qryptonight_7_miner
            return self.qryptonight_7_miner
        else:
            self._current_miner = self.qrandomx_miner
            return self.qrandomx_miner

    def solutionAvailable(self):
        return self._current_miner.solutionAvailable()

    def prepare_next_unmined_block_template(self,
                                            mining_address,
                                            tx_pool,
                                            parent_block: Block,
                                            parent_difficulty,
                                            dev_config: DevConfig):
        miner = self.get_miner(parent_block.block_number + 1, dev_config)
        try:
            logger.debug('Miner-Try - prepare_next_unmined_block_template')
            with self.lock:
                logger.debug('Miner-Locked - prepare_next_unmined_block_template')

                logger.debug('Miner-TryCancel - prepare_next_unmined_block_template')
                miner.cancel()
                logger.debug('Miner-Cancel - prepare_next_unmined_block_template')

                self._mining_block = self.create_block(last_block=parent_block,
                                                       mining_nonce=0,
                                                       tx_pool=tx_pool,
                                                       miner_address=mining_address)

                parent_metadata = self._chain_manager.get_block_metadata(parent_block.headerhash)
                self._measurement = self._chain_manager.get_measurement(dev_config,
                                                                        self._mining_block.timestamp,
                                                                        self._mining_block.prev_headerhash,
                                                                        parent_metadata)

                self._current_difficulty, self._current_target = DifficultyTracker.get(
                    measurement=self._measurement,
                    parent_difficulty=parent_difficulty,
                    dev_config=dev_config)

        except Exception as e:
            logger.warning("Exception in start_mining")
            logger.exception(e)

    def start_mining(self,
                     parent_block: Block,
                     parent_difficulty,
                     dev_config: DevConfig):
        logger.debug('!!! Mine #{} | {} ({}) | {} -> {} | {} '.format(
            self._mining_block.block_number,
            self._measurement, self._mining_block.timestamp - parent_block.timestamp,
            UInt256ToString(parent_difficulty), UInt256ToString(self._current_difficulty),
            bin2hstr(bytearray(self._current_target))
        ))
        logger.debug('!!! Mine #{} | blob: {}'.format(
            self._mining_block.block_number,
            bin2hstr(bytearray(self._mining_block.mining_blob(dev_config)))
        ))
        miner = self.get_miner(parent_block.block_number + 1, dev_config)
        miner.start_mining(self._mining_block, self._current_target, dev_config)

    def create_block(self,
                     last_block,
                     mining_nonce,
                     tx_pool: TransactionPool,
                     miner_address) -> Optional[Block]:
        seed_block = self._chain_manager.get_block_by_number(self._qn.get_seed_height(last_block.block_number + 1))
        dev_config = self._chain_manager.get_config_by_block_number(block_number=last_block.block_number + 1)

        dummy_block = Block.create(dev_config=dev_config,
                                   block_number=last_block.block_number + 1,
                                   prev_headerhash=last_block.headerhash,
                                   prev_timestamp=last_block.timestamp,
                                   transactions=[],
                                   miner_address=miner_address,
                                   seed_height=seed_block.block_number,
                                   seed_hash=seed_block.headerhash)
        dummy_block.set_nonces(dev_config, mining_nonce, 0)

        t_pool2 = tx_pool.transactions

        block_size = dummy_block.size
        block_size_limit = self._chain_manager.get_block_size_limit(last_block, dev_config)

        transactions = []
        state_container = self._chain_manager.new_state_container(set(),
                                                                  last_block.block_number,
                                                                  True,
                                                                  None)
        for tx_set in t_pool2:
            tx = tx_set[1].transaction

            # Skip Transactions for later, which doesn't fit into block
            if block_size + tx.size + dev_config.tx_extra_overhead > block_size_limit:
                continue

            if not self._chain_manager.update_state_container(tx, state_container):
                logger.error("[create_block] Error updating state_container")
                return None

            if not tx.validate_all(state_container, check_nonce=False):
                if not state_container.revert_update():
                    return None
                tx_pool.remove_tx_from_pool(tx)
                continue
            if not self._chain_manager.apply_txn(tx, state_container):
                logger.error("[create_block] Failed to apply txn")
                if not state_container.revert_update():
                    return None
                continue

            addr_from_pk_state = state_container.addresses_state[tx.addr_from]
            addr_from_pk = Transaction.get_slave(tx)
            if addr_from_pk:
                addr_from_pk_state = state_container.addresses_state[addr_from_pk]

            tx._data.nonce = addr_from_pk_state.nonce
            block_size += tx.size + dev_config.tx_extra_overhead
            transactions.append(tx)

        block = Block.create(dev_config=dev_config,
                             block_number=last_block.block_number + 1,
                             prev_headerhash=last_block.headerhash,
                             prev_timestamp=last_block.timestamp,
                             transactions=transactions,
                             miner_address=miner_address,
                             seed_height=seed_block.block_number,
                             seed_hash=seed_block.headerhash)

        return block

    def get_block_to_mine(self,
                          wallet_address,
                          tx_pool,
                          last_block,
                          last_block_difficulty) -> list:
        dev_config = self._chain_manager.get_config_by_block_number(last_block.block_number + 1)
        try:
            mining_address = bytes(hstr2bin(wallet_address[1:].decode()))

            if not OptimizedAddressState.address_is_valid(mining_address):
                raise ValueError("[get_block_to_mine] Invalid Wallet Address %s", wallet_address)
        except Exception as e:
            raise ValueError("Error while decoding wallet address %s", e)

        if self._mining_block:
            if last_block.headerhash == self._mining_block.prev_headerhash:
                if self._mining_block.transactions[0].coinbase.addr_to == mining_address:
                    return [bin2hstr(self._mining_block.mining_blob(dev_config)),
                            int(bin2hstr(self._current_difficulty), 16)]
                else:
                    self._mining_block.update_mining_address(dev_config, mining_address)  # Updates only Miner Address

        self.prepare_next_unmined_block_template(mining_address,
                                                 tx_pool,
                                                 last_block,
                                                 last_block_difficulty,
                                                 dev_config=dev_config)

        return [bin2hstr(self._mining_block.mining_blob(dev_config)),
                int(bin2hstr(self._current_difficulty), 16)]

    def submit_mined_block(self, blob: bytes) -> bool:
        dev_config = self._chain_manager.get_config_by_block_number(self._mining_block.block_number - 1)
        if not self._mining_block.verify_blob(blob, dev_config):
            return False

        blockheader = copy.deepcopy(self._mining_block.blockheader)
        blockheader.set_mining_nonce_from_blob(blob, dev_config)

        dev_config = self._chain_manager.get_config_by_block_number(blockheader.block_number)

        if not self._chain_manager.validate_mining_nonce(blockheader, dev_config):
            return False

        self._mining_block.set_nonces(dev_config,
                                      blockheader.mining_nonce,
                                      blockheader.extra_nonce)
        cloned_block = copy.deepcopy(self._mining_block)
        return self._pre_block_logic(cloned_block)

    def cancel(self):
        self.qryptonight_7_miner.cancel()
        self.qrandomx_miner.cancel()
